/*
 * Copyright 1993-2007 NVIDIA Corporation.  All rights reserved.
 *
 * NOTICE TO USER:
 *
 * This source code is subject to NVIDIA ownership rights under U.S. and
 * international Copyright laws.  Users and possessors of this source code
 * are hereby granted a nonexclusive, royalty-free license to use this code
 * in individual and commercial software.
 *
 * NVIDIA MAKES NO REPRESENTATION ABOUT THE SUITABILITY OF THIS SOURCE
 * CODE FOR ANY PURPOSE.  IT IS PROVIDED "AS IS" WITHOUT EXPRESS OR
 * IMPLIED WARRANTY OF ANY KIND.  NVIDIA DISCLAIMS ALL WARRANTIES WITH
 * REGARD TO THIS SOURCE CODE, INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY, NONINFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE.
 * IN NO EVENT SHALL NVIDIA BE LIABLE FOR ANY SPECIAL, INDIRECT, INCIDENTAL,
 * OR CONSEQUENTIAL DAMAGES, OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
 * OF USE, DATA OR PROFITS,  WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
 * OR OTHER TORTIOUS ACTION,  ARISING OUT OF OR IN CONNECTION WITH THE USE
 * OR PERFORMANCE OF THIS SOURCE CODE.
 *
 * U.S. Government End Users.   This source code is a "commercial item" as
 * that term is defined at  48 C.F.R. 2.101 (OCT 1995), consisting  of
 * "commercial computer  software"  and "commercial computer software
 * documentation" as such terms are  used in 48 C.F.R. 12.212 (SEPT 1995)
 * and is provided to the U.S. Government only as a commercial end item.
 * Consistent with 48 C.F.R.12.212 and 48 C.F.R. 227.7202-1 through
 * 227.7202-4 (JUNE 1995), all U.S. Government End Users acquire the
 * source code with only those rights set forth herein.
 *
 * Any use of this source code in individual and commercial software must
 * include, in the user documentation and internal comments to the code,
 * the above Disclaimer and U.S. Government End Users Notice.
 */

/*
 * This sample implements Mersenne Twister random number generator 
 * and Cartesian Box-Muller transformation on the GPU.
 * See supplied whitepaper for more explanations.
 */



#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <string.h>
#include <cutil.h>
#include <math.h>
#include "MersenneTwister.h"



///////////////////////////////////////////////////////////////////////////////
// Common host and device function 
///////////////////////////////////////////////////////////////////////////////
//ceil(a / b)
extern "C" int iDivUp(int a, int b){
    return ((a % b) != 0) ? (a / b + 1) : (a / b);
}

//floor(a / b)
extern "C" int iDivDown(int a, int b){
    return a / b;
}

//Align a to nearest higher multiple of b
extern "C" int iAlignUp(int a, int b){
    return ((a % b) != 0) ?  (a - a % b + b) : a;
}

//Align a to nearest lower multiple of b
extern "C" int iAlignDown(int a, int b){
    return a - a % b;
}



///////////////////////////////////////////////////////////////////////////////
// Reference MT front-end and Box-Muller transform
///////////////////////////////////////////////////////////////////////////////
extern "C" void initMTRef(const char *fname);
extern "C" void RandomRef(float *h_Random, int NPerRng, unsigned int seed);
extern "C" void BoxMullerRef(float *h_Random, int NPerRng);

void initializeRNG(char **dir); 
__global__ void sigmoidF ( float* g_odata, float* g_idata);
__global__ void sigmoidACu (float* g_odata, float* g_idata);
float cpu_sigmoid (float value);

///////////////////////////////////////////////////////////////////////////////
// Fast GPU random number generator and Box-Muller transform
///////////////////////////////////////////////////////////////////////////////
#include "MersenneTwister_kernel.cu"



///////////////////////////////////////////////////////////////////////////////
// Data configuration
///////////////////////////////////////////////////////////////////////////////
const int    PATH_N = 24000000;
const int N_PER_RNG = iAlignUp(iDivUp(PATH_N, MT_RNG_COUNT), 2);
const int    RAND_N = MT_RNG_COUNT * N_PER_RNG;

const unsigned int SEED = 777; //change this part to - time(NULL)



//#define DO_BOXMULLER //to run Box-Muller transformer to turn [0,1] to a normal distribution (Gaussian probability curve)
///////////////////////////////////////////////////////////////////////////////
// Main program
///////////////////////////////////////////////////////////////////////////////
int main(int argc, char **argv){
    float
        *d_Rand,
        *sig_d_Rand,
        *sig2_d_Rand;

    float
        *h_RandCPU,
        *h_RandGPU,
        *sig_h_Rand,
        *sig2_h_Rand;

    double
        rCPU, rGPU, delta, sum_delta, max_delta, sum_ref, L1norm, gpuTime, cpuTime;
	
	double
        sigCPU, sigGPU, sig_delta, sig_sum_delta, sig_max_delta, sig_sum_ref;

    int i, j;
    unsigned int hTimer;
	
	int *h_HID, *d_HID;
	
    CUT_DEVICE_INIT(argc, argv);
    CUT_SAFE_CALL( cutCreateTimer(&hTimer) );
	
    printf("Initializing data for %i samples...\n", PATH_N);
        h_RandCPU  = (float *)malloc(RAND_N * sizeof(float));
        h_RandGPU  = (float *)malloc(RAND_N * sizeof(float));
        sig_h_Rand  = (float *)malloc(RAND_N * sizeof(float));
        sig2_h_Rand  = (float *)malloc(RAND_N * sizeof(float));
		h_HID = (int *)malloc(RAND_N*sizeof(int));
        CUDA_SAFE_CALL( cudaMalloc((void **)&d_Rand, RAND_N * sizeof(float)) );
		CUDA_SAFE_CALL( cudaMalloc((void **)&d_HID, RAND_N * sizeof(int)) );
        CUDA_SAFE_CALL( cudaMalloc((void **)&sig_d_Rand, RAND_N * sizeof(float)) );
        CUDA_SAFE_CALL( cudaMalloc((void **)&sig2_d_Rand, RAND_N * sizeof(float)) );
/*
    printf("Loading CPU and GPU twisters configurations...\n");
        const char *raw_path = cutFindFilePath("MersenneTwister.raw", argv[0]);
        const char *dat_path = cutFindFilePath("MersenneTwister.dat", argv[0]);
        initMTRef(raw_path);
        loadMTGPU(dat_path);
        seedMTGPU(SEED);
*/
	initializeRNG(argv);
	
    printf("Generating random numbers on GPU...\n");
        CUDA_SAFE_CALL( cudaThreadSynchronize() );
        CUT_SAFE_CALL( cutResetTimer(hTimer) );
        CUT_SAFE_CALL( cutStartTimer(hTimer) );
        RandomGPU<<<32, 128>>>(d_Rand, N_PER_RNG);
        CUT_CHECK_ERROR("RandomGPU() execution failed\n");
        CUDA_SAFE_CALL( cudaThreadSynchronize() );
        CUT_SAFE_CALL( cutStopTimer(hTimer) );
        gpuTime = cutGetTimerValue(hTimer);
    printf("Generated samples : %i \n", RAND_N);
    printf("RandomGPU() time  : %f \n", gpuTime);
    printf("Samples per second: %E \n", RAND_N / (gpuTime * 0.001));

	printf("Generating random numbers on GPU and comparing with sigmoid...\n");
        CUDA_SAFE_CALL( cudaThreadSynchronize() );
        CUT_SAFE_CALL( cutResetTimer(hTimer) );
        CUT_SAFE_CALL( cutStartTimer(hTimer) );
        learnGPU<<<32, 128>>>(d_HID, d_Rand, N_PER_RNG);
        CUT_CHECK_ERROR("learnGPU() execution failed\n");
        CUDA_SAFE_CALL( cudaThreadSynchronize() );
        CUT_SAFE_CALL( cutStopTimer(hTimer) );
        gpuTime = cutGetTimerValue(hTimer);
    printf("Generated samples : %i \n", RAND_N);
    printf("learnGPU() time  : %f \n", gpuTime);
    printf("Samples per second: %E \n", RAND_N / (gpuTime * 0.001));

#ifdef DO_BOXMULLER
    printf("Applying Box-Muller transformation on GPU...\n");
        CUDA_SAFE_CALL( cudaThreadSynchronize() );
        CUT_SAFE_CALL( cutResetTimer(hTimer) );
        CUT_SAFE_CALL( cutStartTimer(hTimer) );
        BoxMullerGPU<<<32, 128>>>(d_Rand, N_PER_RNG);
        CUT_CHECK_ERROR("BoxMullerGPU() execution failed\n");
        CUDA_SAFE_CALL( cudaThreadSynchronize() );
        CUT_SAFE_CALL( cutStopTimer(hTimer) );
        gpuTime = cutGetTimerValue(hTimer);
    printf("Transformed samples : %i \n", RAND_N);
    printf("BoxMullerGPU() time : %f \n", gpuTime);
    printf("Samples per second  : %E \n", RAND_N / (gpuTime * 0.001));
#endif
	
	printf("Inputing random numbers into sigmoid2...\n");
        CUDA_SAFE_CALL( cudaThreadSynchronize() );
        CUT_SAFE_CALL( cutResetTimer(hTimer) );
        CUT_SAFE_CALL( cutStartTimer(hTimer) );
		sigmoidACu<<<46880, 512>>>(sig2_d_Rand, d_Rand);
		CUT_CHECK_ERROR("sigmoidACu() execution failed\n");
        CUDA_SAFE_CALL( cudaThreadSynchronize() );
        CUT_SAFE_CALL( cutStopTimer(hTimer) );
        gpuTime = cutGetTimerValue(hTimer);
    printf("Function samples : %i \n", RAND_N);
    printf("sigmoidACu() time  : %f \n", gpuTime);
    printf("Samples per second: %E \n", RAND_N / (gpuTime * 0.001));
	
	printf("Inputing random numbers into sigmoid...\n");
        CUDA_SAFE_CALL( cudaThreadSynchronize() );
        CUT_SAFE_CALL( cutResetTimer(hTimer) );
        CUT_SAFE_CALL( cutStartTimer(hTimer) );
		sigmoidF<<<46880, 512>>>(sig_d_Rand, d_Rand);
		CUT_CHECK_ERROR("sigmoidF() execution failed\n");
        CUDA_SAFE_CALL( cudaThreadSynchronize() );
        CUT_SAFE_CALL( cutStopTimer(hTimer) );
        gpuTime = cutGetTimerValue(hTimer);
    printf("Function samples : %i \n", RAND_N);
    printf("sigmoidF() time  : %f \n", gpuTime);
    printf("Samples per second: %E \n", RAND_N / (gpuTime * 0.001));
    
    printf("Reading back the results...\n");
        CUDA_SAFE_CALL( cudaMemcpy(h_RandGPU, d_Rand, RAND_N * sizeof(float), cudaMemcpyDeviceToHost) );
	
	printf("Reading back the sig results...\n");
        CUDA_SAFE_CALL( cudaMemcpy(sig_h_Rand, sig_d_Rand, RAND_N * sizeof(float), cudaMemcpyDeviceToHost) );
	
	printf("Reading back the sig2 results...\n");
        CUDA_SAFE_CALL( cudaMemcpy(sig2_h_Rand, sig2_d_Rand, RAND_N * sizeof(float), cudaMemcpyDeviceToHost) );
	
	printf("Reading back the HID results...\n");
        CUDA_SAFE_CALL( cudaMemcpy(h_HID, d_HID, RAND_N * sizeof(int), cudaMemcpyDeviceToHost) );
	
	printf("Reading an rand element ... %f\n", h_RandGPU[0]);
	printf("Reading an sig element ... %f\n", sig_h_Rand[0]);
	printf("Reading an sig2 element ... %f\n", sig2_h_Rand[0]);
	printf("Reading an HID element ... %u\n", h_HID[0]);

    printf("Checking GPU results...\n");
        printf("...generating random numbers on CPU using reference generator\n");
		CUDA_SAFE_CALL( cudaThreadSynchronize() );
        CUT_SAFE_CALL( cutResetTimer(hTimer) );
        CUT_SAFE_CALL( cutStartTimer(hTimer) );
        RandomRef(h_RandCPU, N_PER_RNG, SEED);
		CUDA_SAFE_CALL( cudaThreadSynchronize() );
        CUT_SAFE_CALL( cutStopTimer(hTimer) );
        cpuTime = cutGetTimerValue(hTimer);
    printf("Generated samples : %i \n", RAND_N);
    printf("RandomRef() (CPU) time  : %f \n", cpuTime);
    printf("Samples per second: %E \n", RAND_N / (cpuTime * 0.001));

        #ifdef DO_BOXMULLER
            printf("...applying Box-Muller transformation on CPU\n");
            BoxMullerRef(h_RandCPU, N_PER_RNG);
        #endif

        printf("...comparing the results\n");
        max_delta = 0;
        sum_delta = 0;
        sum_ref   = 0;
		sig_max_delta = 0;
        sig_sum_delta = 0;
        sig_sum_ref   = 0;
		printf("... just for reference %f %f\n%f %f\n%f %f\n%f %f\n",cpu_sigmoid(h_RandCPU[0]), sig_h_Rand[0],cpu_sigmoid(h_RandCPU[1]), sig_h_Rand[4096],cpu_sigmoid(h_RandCPU[2]), sig_h_Rand[8192],cpu_sigmoid(h_RandCPU[3]), sig_h_Rand[12288],cpu_sigmoid(h_RandCPU[4]), sig_h_Rand[16384]);
        for(i = 0; i < MT_RNG_COUNT; i++)
            for(j = 0; j < N_PER_RNG; j++){
                rCPU = h_RandCPU[i * N_PER_RNG + j];
                rGPU = h_RandGPU[i + j * MT_RNG_COUNT];
                delta = fabs(rCPU - rGPU);
				sigCPU = cpu_sigmoid(rCPU);
				sigGPU =  sig_h_Rand[i + j * MT_RNG_COUNT]; //cpu_sigmoid(rGPU); //
				sig_delta = fabs(sigCPU - sigGPU);
                sum_delta += delta;
				sig_sum_delta += sig_delta;
                sum_ref  += fabs(rCPU);
				sig_sum_ref += fabs(sigCPU);
                if(delta >= max_delta) max_delta = delta;
				if(sig_delta >= sig_max_delta) sig_max_delta = sig_delta;
            }
    L1norm = (float)(sum_delta / sum_ref);
    printf("Max absolute error: %E\n", max_delta);
    printf("L1 norm: %E\n", L1norm);
    printf((L1norm < 1e-6) ? "TEST PASSED\n" : "TEST FAILED\n");
	printf("Max sig absolute error: %E\n", sig_max_delta);
    L1norm = (float)(sig_sum_delta / sig_sum_ref);
	printf("L1 sig norm: %E\n", L1norm);	
    printf((L1norm < 1e-6) ? "sig TEST PASSED\n" : "sig TEST FAILED\n");

    printf("Shutting down...\n");
        CUDA_SAFE_CALL( cudaFree(d_Rand) );
		CUDA_SAFE_CALL( cudaFree(sig_d_Rand) );
		CUDA_SAFE_CALL( cudaFree(sig2_d_Rand) );
		CUDA_SAFE_CALL( cudaFree(d_HID) );
        free(h_RandGPU);
        free(h_RandCPU);
		free(sig_h_Rand);
		free(sig2_h_Rand);
		free(h_HID);

    CUT_SAFE_CALL( cutDeleteTimer( hTimer) );

    CUT_EXIT(argc, argv);
}

void initializeRNG(char **dir) {
	printf("Loading CPU and GPU twisters configurations...\n");
        const char *raw_path = cutFindFilePath("MersenneTwister.raw", dir[0]);
        const char *dat_path = cutFindFilePath("MersenneTwister.dat", dir[0]);
		//printf("path: %s\n", raw_path);
        initMTRef(raw_path);
        loadMTGPU(dat_path);
        seedMTGPU(SEED);
}

float cpu_sigmoid (float value) {
	return 1/(1+expf(-value));
}

/*FLOATS*/
__global__ void sigmoidF ( float* g_odata, float* g_idata) {
   unsigned int x = blockIdx.x*blockDim.x + threadIdx.x;
   g_odata[x] = 1/(1+__expf(-g_idata[x]));
}

/* Sigmoid curve split into*/
__global__ void sigmoidACu (float* g_odata, float* g_idata) {
    unsigned int x = blockIdx.x*blockDim.x + threadIdx.x;
    int m = 32; float b = 1.0f; //in>8
    signed int in = (signed int)(g_idata[x]*134217728); //
    // out = in<<m + b, find m and b
    if      (in<-8<<27) { m = 32; b = 0.0000f; }
    else if (in<-4<<27) { m =  6; b = 0.1250f; }
    else if (in<-2<<27) { m =  5; b = 0.1875f; }
    else if (in<-1<<27) { m =  3; b = 0.3750f; }
    else if (in< 1<<27) { m =  2; b = 0.5000f; }
    else if (in< 2<<27) { m =  3; b = 0.6250f; }
    else if (in< 4<<27) { m =  5; b = 0.8125f; }
    else if (in< 8<<27) { m =  6; b = 0.8750f; }
    g_odata[x] = (float)(in>>m)/134217728 + b;
}
 
