#include "MersenneTwister.h"
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "dci.h"
#include <cutil.h>
#include "sigmoid.cu"

static mt_struct MT[MT_RNG_COUNT];
static uint32_t state[MT_NN];

extern "C"{
    //ceil(a / b)
    int iDivUp(int a, int b){
        return ((a % b) != 0) ? (a / b + 1) : (a / b);
    }

    //floor(a / b)
    int iDivDown(int a, int b){
        return a / b;
    }

    //Align a to nearest higher multiple of b
    int iAlignUp(int a, int b){
        return ((a % b) != 0) ?  (a - a % b + b) : a;
    }

    //Align a to nearest lower multiple of b
    int iAlignDown(int a, int b){
        return a - a % b;
    }

    void initMTRef(const char *fname){
    
        FILE *fd = fopen(fname, "rb");
        if(!fd){
            printf("initMTRef(): failed to open %s\n", fname);
            printf("TEST FAILED\n");
            exit(0);
        }
    
        for (int i = 0; i < MT_RNG_COUNT; i++){
            //Inline structure size for compatibility,
            //since pointer types are 8-byte on 64-bit systems (unused *state variable)
            if( !fread(MT + i, 16 /* sizeof(mt_struct) */ * sizeof(int), 1, fd) ){
                printf("initMTRef(): failed to load %s\n", fname);
                printf("TEST FAILED\n");
                exit(0);
            }
        }
    
        fclose(fd);
    }   
}

// Data configuration
const int    PATH_N    = 262144;
const int    N_PER_RNG = iAlignUp(iDivUp(PATH_N, MT_RNG_COUNT), 2);
const int    RAND_N    = MT_RNG_COUNT * N_PER_RNG;

const unsigned int SEED = 777; //change this part to - time(NULL)

__device__ static mt_struct_stripped ds_MT[MT_RNG_COUNT];
static mt_struct_stripped h_MT[MT_RNG_COUNT];

//Load twister configurations
void loadMTGPU(const char *fname){
    FILE *fd = fopen(fname, "rb");
    if(!fd){
        printf("initMTGPU(): failed to open %s\n", fname);
        printf("TEST FAILED\n");
        exit(0);
    }
    if( !fread(h_MT, sizeof(h_MT), 1, fd) ){
        printf("initMTGPU(): failed to load %s\n", fname);
        printf("TEST FAILED\n");
        exit(0);
    }
    fclose(fd);
}

//Initialize/seed twister for current GPU context
void seedMTGPU(unsigned int seed){
    int i;
    //Need to be thread-safe
    mt_struct_stripped *MT = (mt_struct_stripped *)malloc(MT_RNG_COUNT * sizeof(mt_struct_stripped));

    for(i = 0; i < MT_RNG_COUNT; i++){
        MT[i]      = h_MT[i];
        MT[i].seed = seed;
    }
    CUDA_SAFE_CALL( cudaMemcpyToSymbol(ds_MT, MT, sizeof(h_MT)) );

    free(MT);
}

//Executable and main .cu file must have the same name!
void initializeRNG() {
	printf("Loading CPU and GPU twisters configurations...\n");
    initMTRef("MersenneTwister.raw");
    loadMTGPU("MersenneTwister.dat");
    seedMTGPU(SEED);
}

__global__ void GPUtransfer(int *o_matrix, float *d_energy, int NPerRng){
    const int      tid = blockDim.x * blockIdx.x + threadIdx.x;
    const int THREAD_N = blockDim.x * gridDim.x;

    int iState, iState1, iStateM, iOut;
    unsigned int mti, mti1, mtiM, x;
    unsigned int mt[MT_NN];

    for(int iRng = tid; iRng < MT_RNG_COUNT; iRng += THREAD_N){
        //Load bit-vector Mersenne Twister parameters
        mt_struct_stripped config = ds_MT[iRng];

        //Initialize current state
        mt[0] = config.seed;
        for(iState = 1; iState < MT_NN; iState++)
            mt[iState] = (1812433253U * (mt[iState - 1] ^ (mt[iState - 1] >> 30)) + iState) & MT_WMASK;

        iState = 0;
        mti1 = mt[0];
        for(iOut = 0; iOut < NPerRng; iOut++){
            //iState1 = (iState +     1) % MT_NN
            //iStateM = (iState + MT_MM) % MT_NN
            iState1 = iState + 1;
            iStateM = iState + MT_MM;
            if(iState1 >= MT_NN) iState1 -= MT_NN;
            if(iStateM >= MT_NN) iStateM -= MT_NN;
            mti  = mti1;
            mti1 = mt[iState1];
            mtiM = mt[iStateM];

            x    = (mti & MT_UMASK) | (mti1 & MT_LMASK);
            x    =  mtiM ^ (x >> 1) ^ ((x & 1) ? config.matrix_a : 0);
            mt[iState] = x;
            iState = iState1;

            //Tempering transformation
            x ^= (x >> MT_SHIFT0);
            x ^= (x << MT_SHIFTB) & config.mask_b;
            x ^= (x << MT_SHIFTC) & config.mask_c;
            x ^= (x >> MT_SHIFT1);

            //Convert to (0, 1] float and write to global memory
            float random = ((float)x + 1.0f) / 4294967296.0f;
            o_matrix[iRng + iOut * MT_RNG_COUNT] = sigmoidF(d_energy[iRng + iOut * MT_RNG_COUNT]) > random;
        }
    }
}
