/* Matrix multiplication: C = A * B.
 * Device code.
 */

#ifndef _MATRIX_KERNEL_H_
#define _MATRIX_KERNEL_H_
#define BLOCK_SIZE 16
#include <stdio.h>

////////////////////////////////////////////////////////////////////////////////
//! Matrix multiplication on the device: C = A * B
//! wA is A's width and wB is B's width
////////////////////////////////////////////////////////////////////////////////
template <int wA, int wB, typename TA, typename TB>
__global__ void matrixMult(TB* C, TA* A, TB* B){
    // Block index
    int bx = blockIdx.x;
    int by = blockIdx.y;

    // Thread index
    int tx = threadIdx.x;
    int ty = threadIdx.y;

    // Index of the first sub-matrix of A processed by the block
    int aBegin = wA * BLOCK_SIZE * by;

    // Index of the last sub-matrix of A processed by the block
    int aEnd   = aBegin + wA - 1;

    // Step size used to iterate through the sub-matrices of A
    int aStep  = BLOCK_SIZE;

    // Index of the first sub-matrix of B processed by the block
    int bBegin = BLOCK_SIZE * bx;

    // Step size used to iterate through the sub-matrices of B
    int bStep  = BLOCK_SIZE * wB;

    // Csub is used to store the element of the block sub-matrix
    // that is computed by the thread
    TB Csub = 0;

    // Loop over all the sub-matrices of A and B
    // required to compute the block sub-matrix
    for (int a = aBegin, b = bBegin;
             a <= aEnd;
             a += aStep, b += bStep) {

        // Declaration of the shared memory array As used to
        // store the sub-matrix of A
        __shared__ TA As[BLOCK_SIZE][BLOCK_SIZE];

        // Declaration of the shared memory array Bs used to
        // store the sub-matrix of B
        __shared__ TB Bs[BLOCK_SIZE][BLOCK_SIZE];

        // Load the matrices from device memory
        // to shared memory; each thread loads
        // one element of each matrix
        As[ty][tx] = A[a + wA * ty + tx];
        Bs[ty][tx] = B[b + wB * ty + tx];

        // Synchronize to make sure the matrices are loaded
        __syncthreads();

        // Multiply the two matrices together;
        // each thread computes one element
        // of the block sub-matrix
        for (int k = 0; k < BLOCK_SIZE; ++k)
            Csub += (TB)As[ty][k] * Bs[k][tx]; //(TB)(As[ty][k] * Bs[k][tx]); //(TB)As[ty][k] * Bs[k][tx]

        // Synchronize to make sure that the preceding
        // computation is done before loading two new
        // sub-matrices of A and B in the next iteration
        __syncthreads();
    }

    // Write the block sub-matrix to device memory;
    // each thread writes one element
    int c = wB * BLOCK_SIZE * by + BLOCK_SIZE * bx;
    C[c + wB * ty + tx] = Csub;
}

template <int hA, int wA, typename T>
__global__ void matrixTranspose (T *B, T *A){
    __shared__ T block [(BLOCK_SIZE+1)*BLOCK_SIZE];

    unsigned int xBlock = blockDim.x*blockIdx.x;
    unsigned int yBlock = blockDim.y*blockIdx.y;
    unsigned int xIndex = xBlock + threadIdx.x;
    unsigned int yIndex = yBlock + threadIdx.y;
    unsigned int index_out, index_transpose;

    if ((xIndex < wA) && (yIndex < hA))
    {
        unsigned int index_in = wA*yIndex + xIndex;
        unsigned int index_block = threadIdx.y*(BLOCK_SIZE+1) + threadIdx.x;
        block[index_block] = A[index_in];
        index_transpose = threadIdx.x*(BLOCK_SIZE+1) + threadIdx.y;
        index_out = hA*(xBlock + threadIdx.y) + yBlock + threadIdx.x;
    }

    __syncthreads();

    if ((xIndex < wA) && (yIndex < hA))
    {
        B[index_out] = block[index_transpose];
    }
}

//__global__ void matrixMac(float *C, int *A, int *B, int hA, int wA, float epsilon)
template <int hA, int wA, typename T>
__global__ void matrixMac  (float *C, T *A, T *B, float epsilon){
    uint x = blockIdx.x * blockDim.x + threadIdx.x;
    uint y = blockIdx.y * blockDim.y + threadIdx.y;

    if ((x < wA) && (y < hA))
    {
        C[y*wA + x] = C[y*wA + x] + epsilon*(A[y*wA + x] - B[y*wA + x]);
    }
}

/////////////////////////////////////////////////////////////////
// Templated interface
/////////////////////////////////////////////////////////////////
// Compute matrix multiply
/*
template <int hA, int wA, int wB, typename TA, typename TB>
void matrix_mult (TB *C, TA *A, TB *B) {
    dim3 threads(BLOCK_SIZE, BLOCK_SIZE);
    //dim3 grid(wB/threads.x, hA/threads.y); //dim3 grid(hA/BLOCK_SIZE, wB/BLOCK_SIZE);
    dim3 grid(hA/BLOCK_SIZE, wB/BLOCK_SIZE);
    matrixMult<wA, wB, TA, TB> <<< grid, threads >>> (C, A, B);
}
*/
// Compute matrix tranpose

template <int hA, int wA, typename T>
void matrix_transpose(T *B, T *A) {
    dim3 threads(BLOCK_SIZE, BLOCK_SIZE);
    //dim3 grid(wA/threads.x, hA/threads.y);
    dim3 grid(wA/BLOCK_SIZE, hA/BLOCK_SIZE);
    //matrixTranspose<hA, wA, T> <<< grid, threads, (BLOCK_SIZE+1)*BLOCK_SIZE*sizeof(T) >>>(B, A);
    matrixTranspose<hA, wA, T> <<< grid, threads >>>(B, A);
}

#endif // #ifndef _MATRIX_KERNEL_H_
